#include "matrix.h"
#include <math.h>
#include <stdio.h>
#include "font.h"
#include "vector.h"

__declspec(align(16)) struct DPoint
{
  float x, y, z, w, r, g, b, a, u ,v;
  float normal[3]; // per-vertex normal
};

__declspec(align(16)) struct DTriangle
{
  DPoint p[3];
  float normal[3], zero; // per-face normal
  bool Fill;
};

__declspec(align(16)) struct DMatrix
{
  float d[16];
};

static DTriangle Triangles[2048];
static DMatrix   Matrixes[2048];
static DMatrix   *CurrentMatrix = Matrixes;

static int TriangleCount;

static DMatrix   IdentityMatrix =
{
  {
    1.0f, 0.0f, 0.0f, 0.0f,
    0.0f, 1.0f, 0.0f, 0.0f,
    0.0f, 0.0f, 1.0f, 0.0f,
    0.0f, 0.0f, 0.0f, 1.0f,
  }
};

//extern unsigned char font[][8];
extern unsigned int *BackBuffer;

enum
{
  PNT_X, // 0
  PNT_Y, // 1
  PNT_Z, // 2
  PNT_W, // 3
  PNT_R, // 4
  PNT_G, // 5
  PNT_B, // 6
  PNT_A, // 7
  PNT_U, // 8
  PNT_V, // 9
};

static bool G3D_ModelFill;

typedef void (*VertexShaderProc)(DPoint *p);

void VShaderIdentity(DPoint *p)
{
  return;
}

__declspec(naked) void VShaderFlat(DPoint *p)
{
  __asm
  {
    mov eax, [esp+4]                 // p
    mov dword ptr [eax+8], 42C80000h // p->z = 100.0f
    retn
  };
}

VertexShaderProc VShader;

typedef unsigned int (*PixelShaderProc)(float rgba[4], int x, int y, float z, float u, float v, float nx, float ny, float nz);

unsigned int PShaderColor(float rgba[4], int x, int y, float z, float u, float v, float nx, float ny, float nz);
unsigned int PShaderDist(float rgba[4], int x, int y, float z, float u, float v, float nx, float ny, float nz);
unsigned int PShaderPos(float rgba[4], int x, int y, float z, float u, float v, float nx, float ny, float nz);
PixelShaderProc PShader;
typedef float ScanBuffer[480];

static void Scan(ScanBuffer buffers[10], DPoint *a, DPoint *b)
{
  int k;
  float *ca = (float*)a;
  float *cb = (float*)b;
  float ydiff = b->y - a->y + 1;

  for(k = 0; k < 10; k++)
  {
    float cnt = ca[k];
    float m = (cb[k] - ca[k]) / ydiff;

    int i;
    for(i = (int)a->y; i <= (int)b->y; i++)
    {
      buffers[k][i] = cnt;
      cnt += m;
    }
  } 
}

static void Swap(DPoint **a, DPoint **b)
{
  DPoint *c;
  c = *a;
  *a = *b;
  *b = c;
}

static unsigned int ColorFloatToUint(float *color)
{
  unsigned int r,g,b,a;
  r = (unsigned int)color[0]; if(r > 0xff) r = 0xff;
  g = (unsigned int)color[1]; if(g > 0xff) g = 0xff;
  b = (unsigned int)color[2]; if(b > 0xff) b = 0xff;
  a = (unsigned int)color[3]; if(a > 0xff) a = 0xff;

  return b | (g << 8 ) | (r << 16) | (a << 24);
}

static unsigned int BindColors(unsigned int aa, unsigned int bb)
{
  unsigned int r,g,b;
  unsigned int alpha = aa >> 24;
  b = ((aa & 0xff) * alpha + (bb & 0xff) * (255 - alpha)) >> 8;
  g = (((aa >>  8) & 0xff) * alpha + ((bb >>  8) & 0xff) * (255 - alpha)) >> 8;
  r = (((aa >> 16) & 0xff) * alpha + ((bb >> 16) & 0xff) * (255 - alpha)) >> 8;
  return b | (g << 8 ) | (r << 16) | (alpha << 24);
}

unsigned int PShaderColor(float rgba[4], int x, int y, float z, float u, float v, float nx, float ny, float nz)
{
  float my_rgba[4] = {rgba[0], rgba[1], rgba[2], rgba[3]};
  float r_x = (float)(x - 320) / 320.0f * z;
  float r_y = (float)(y - 240) / 320.0f * z;
  DVector vn = {nx, ny, nz, 0.0f};
  DVector vp = {-r_x, -r_y, -z, 0.0f};
  v_normalize(&vp);
  
  float diff = fabs(v_dot(&vn, &vp)) + 0.1f;
  float spec = pow(diff, 128.0f);
  if(diff + spec > 0.0f)
  {
    if(spec > 0.8f)
    {
      return 0xffffffff;
    }
    else
    {
      my_rgba[0] = my_rgba[0] * diff + spec * 255.0f;
      my_rgba[1] = my_rgba[1] * diff + spec * 255.0f;
      my_rgba[2] = my_rgba[2] * diff + spec * 255.0f;
    }
  }
  else
  {
    return 0;
  }

  return ColorFloatToUint(my_rgba);  
}

unsigned int PShaderPos(float rgba[4], int x, int y, float z, float u, float v, float nx, float ny, float nz)
{
  float myrgba[4];
  myrgba[3] = 255.0f;
  myrgba[0] = (float)x / 640.0f * 255.0f;
  myrgba[2] = (float)y / 480.0f * 255.0f;
  myrgba[1] = 0.0f;;

  return ColorFloatToUint(myrgba);  
}

float max_dist = 100.0f;
static int my_rand_seed = 0x1234567;

int my_rand()
{
  return my_rand_seed = (my_rand_seed << 1) ^ (my_rand_seed >> 2) ^ (my_rand_seed + 123);
}

unsigned int PShaderDist(float rgba[4], int x, int y, float z, float u, float v, float nx, float ny, float nz)
{
  unsigned int c = (unsigned int)(((max_dist - z)/max_dist) * 255.0f);
  int d = (int)c + (my_rand() & 0x0f);
     
  c = (unsigned int)d;
  c |= (c << 8) | (c << 16) | 0xff000000;

  return c;  
}


static void SetPixel(unsigned int *Buffer, int x, int y, float z, unsigned int color)
{
  Buffer[x + y * 640] = BindColors(color, Buffer[x + y * 640]);
}
 
static void DrawTriangle(unsigned int *Buffer, DTriangle *tr)
{
  ScanBuffer *scan_left, *scan_right;
  static ScanBuffer scan_a[10], scan_b[10];

  enum { MIN, MID, MAX };
  DPoint *ver[3];
  ver[0] = &tr->p[0];
  ver[1] = &tr->p[1];
  ver[2] = &tr->p[2];

  if(ver[0]->y > ver[1]->y) Swap(&ver[0], &ver[1]);
  if(ver[1]->y > ver[2]->y) Swap(&ver[1], &ver[2]);
  if(ver[0]->y > ver[1]->y) Swap(&ver[0], &ver[1]);

  Scan(scan_a, ver[MIN], ver[MAX]);
  Scan(scan_b, ver[MIN], ver[MID]);
  Scan(scan_b, ver[MID], ver[MAX]);

  if(scan_a[PNT_X][(int)ver[MID]->y] > ver[MID]->x)
  {
    // pointer to left
    scan_left = scan_b;
    scan_right = scan_a;
  }
  else
  {
    // pointer to right
    scan_left = scan_a;
    scan_right = scan_b;
  }

  int i, j, k;
  for(i = (int)ver[MIN]->y; i <= (int)ver[MAX]->y; i++)
  {
    if(tr->Fill || i == (int)ver[MIN]->y || i == (int)ver[MAX]->y)
    {
      float z, z_m;
      float color[4], color_m[4], x_diff;
      x_diff = scan_right[PNT_X][i] - scan_left[PNT_X][i] + 1;

      for(k = 0; k < 4; k++)
      {
        color[k] = scan_left[PNT_R+k][i];
        color_m[k] = (scan_right[PNT_R+k][i] - scan_left[PNT_R+k][i]) / x_diff;
      }
      z = scan_left[PNT_Z][i];
      z_m = (scan_right[PNT_Z][i] - scan_left[PNT_Z][i]) / x_diff;

      for(j = (int)scan_left[PNT_X][i]; j <= (int)scan_right[PNT_X][i]; j++)
      {
        SetPixel(Buffer, j, i, z,
	    PShader(color, j, i, z, 1.0f, 1.0f, tr->normal[0], tr->normal[1], tr->normal[2]));

        color[0] += color_m[0];
        color[1] += color_m[1];
        color[2] += color_m[2];
        color[3] += color_m[3];
	z += z_m;
      }
    }
  }
}

static void ClearBuffer(unsigned int *b, int size)
{
  while(size--) *b++ = 0;
}

static void CopyBuffer(unsigned int *a, const unsigned int *b, int size)
{
  while(size--) *a++ = *b++;
}

#define CopyBuffer(a,b,c) CopyBuffer((unsigned int*)(a), (const unsigned int*)(b), (c))
#define ClearBuffer(a,c) ClearBuffer((unsigned int*)(a), (c))

static void G3D_Translate(float x, float y, float z)
{
  static DMatrix temp = IdentityMatrix;
  temp.d[3+0] = x;
  temp.d[3+4] = y;
  temp.d[3+8] = z;
  MMMul(CurrentMatrix, &temp, CurrentMatrix+1);
  CurrentMatrix++;
}

static void G3D_Scale(float x, float y, float z)
{
  static DMatrix temp = IdentityMatrix;
  temp.d[0+0] = x;
  temp.d[1+4] = y;
  temp.d[2+8] = z;
  MMMul(CurrentMatrix, &temp, CurrentMatrix+1);
  CurrentMatrix++;
}

static void G3D_RotateAxisX(float r)
{
  static DMatrix temp = IdentityMatrix;
  temp.d[1+4] = cosf(r);
  temp.d[2+4] = -sinf(r);
  temp.d[1+8] = sinf(r);
  temp.d[2+8] = cosf(r);
  MMMul(CurrentMatrix, &temp, CurrentMatrix+1);
  CurrentMatrix++; 
}

static void G3D_RotateAxisY(float r)
{
  static DMatrix temp = IdentityMatrix;
  temp.d[0+0] = cosf(r);
  temp.d[2+0] = sinf(r);
  temp.d[0+8] = -sinf(r);
  temp.d[2+8] = cosf(r);
  MMMul(CurrentMatrix, &temp, CurrentMatrix+1);
  CurrentMatrix++;
}

static void G3D_AddTriangle(DTriangle *tr)
{
  CopyBuffer(&Triangles[TriangleCount], tr, sizeof(*tr)/4);
  MVMul(CurrentMatrix, &tr->p[0], &Triangles[TriangleCount].p[0]); 
  MVMul(CurrentMatrix, &tr->p[1], &Triangles[TriangleCount].p[1]); 
  MVMul(CurrentMatrix, &tr->p[2], &Triangles[TriangleCount].p[2]); 
  MVMul(CurrentMatrix, &tr->normal, &Triangles[TriangleCount].normal); 

  Triangles[TriangleCount].Fill = G3D_ModelFill;
  TriangleCount++;
}

static void G3D_Reset()
{
  TriangleCount = 0;
  CurrentMatrix  = &Matrixes[0];  
  *CurrentMatrix = IdentityMatrix;
}

static void G3D_Render(unsigned int *Buffer)
{
  int i, j;
  for(j = 0; j < TriangleCount; j++)
  {
    for(i = 0; i < 3; i++)
    {
      VShader(&Triangles[j].p[i]);
      Triangles[j].p[i].x = 320.0f + 320.0f * (Triangles[j].p[i].x / Triangles[j].p[i].z);
      Triangles[j].p[i].y = 240.0f + 320.0f * (Triangles[j].p[i].y / Triangles[j].p[i].z);
    }
    DrawTriangle(Buffer, &Triangles[j]);
  }
}


static DPoint QuadVertex[8] = {
  {  1.0,  1.0,  1.0,  1.0  }, // RLB
  {  1.0,  1.0, -1.0,  1.0  }, // RLF
  {  1.0, -1.0,  1.0,  1.0  }, // RUB
  {  1.0, -1.0, -1.0,  1.0  }, // RUF
  { -1.0,  1.0,  1.0,  1.0  }, // LLB
  { -1.0,  1.0, -1.0,  1.0  }, // LLF
  { -1.0, -1.0,  1.0,  1.0  }, // LUB
  { -1.0, -1.0, -1.0,  1.0  }  // LUF
};

static int QuadFace[6][4] = {
  { 0, 1, 2, 3 },
  { 4, 5, 6, 7 },
  { 0, 1, 4, 5 },
  { 2, 3, 6, 7 },
  { 0, 2, 4, 6 },
  { 1, 3, 5, 7 }
};

static void AddQuad()
{
  int i;
  static DTriangle temp;
  
  for(i = 0; i < 6; i++)
  {
    CopyBuffer(&temp.p[0], &QuadVertex[QuadFace[i][0]], 4);
    CopyBuffer(&temp.p[1], &QuadVertex[QuadFace[i][1]], 4);
    CopyBuffer(&temp.p[2], &QuadVertex[QuadFace[i][2]], 4);
    G3D_AddTriangle(&temp);

    CopyBuffer(&temp.p[0], &QuadVertex[QuadFace[i][3]], 4);
    G3D_AddTriangle(&temp);
  }
}

static void DrawString3D(const char *string, int len)
{
  int j, k;
  bool skip[8];

  while(len--)
  {  
    for(j = 0; j < 8; j++)
    {
      unsigned char b = font[*string][j];
      for(k = 0; k < 8; k++)
      {
       	if(b & 0x80) AddQuad();
	b <<= 1;
	G3D_Translate(2.0f, 0.0f, 0.0f);
      }
      G3D_Translate(-16.0f, 2.0f, 0.0f);
    }
    G3D_Translate(18.0f, -16.0f, 0.0f);

    string++;
  }
}

#include "palette.h"

unsigned int *flame_pal = (unsigned int*)flame_pallete_char;

static DPoint    
  Amethyst_PointTop    = { 0.3f, -3.0f, 0.5f, 1.0f, 255.0f, 255.0f, 255.0f, 128.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f },
  Amethyst_PointMiddle[6],
  Amethyst_PointMiddleB[6];

static DTriangle Amethyst_Model[6+12]; // top, middle, bottom

void TriangleCalcNormal(DTriangle *tr)
{
  static DVector v1, v2, v3;

  // go go static code
  CopyBuffer(&v1, &tr->p[0], 3);
  CopyBuffer(&v2, &tr->p[1], 3);
  CopyBuffer(&v3, &tr->p[2], 3);
  v_sub(&v1, &v2);
  v_sub(&v2, &v3);
  
  tr->normal[0] = v1.y * v2.z - v1.z * v2.y;
  tr->normal[1] = v1.z * v2.x - v1.x * v2.z;
  tr->normal[2] = v1.x * v2.y - v1.y * v2.x;
  tr->zero = 0;
  v_normalize((DVector*)&tr->normal);
}

void InitAmethyst()
{
  int i;
  static float ang;
  for(i = 0; i < 6; i++)
  {
    ang += (2.0f * 3.14f / 6.0f);
    Amethyst_PointMiddle[i].x = cosf(ang);
    Amethyst_PointMiddle[i].y = -1.0f;;
    Amethyst_PointMiddle[i].z = sinf(ang);
    Amethyst_PointMiddle[i].w = 1.0f;
    Amethyst_PointMiddle[i].r = 255.0f;
    Amethyst_PointMiddle[i].b = 255.0f;
    Amethyst_PointMiddle[i].a = 127.0f;

    CopyBuffer(&Amethyst_PointMiddleB[i], &Amethyst_PointMiddle[i], 13);
    Amethyst_PointMiddleB[i].y += 4.0f;
    Amethyst_PointMiddleB[i].r = 0.0f;
    Amethyst_PointMiddleB[i].b = 0.0f;
    Amethyst_PointMiddleB[i].x *= 0.5f;
    Amethyst_PointMiddleB[i].z *= 0.5f;
  }

  for(i = 0; i < 6; i++)
  {
    CopyBuffer(&Amethyst_Model[i].p[1], &Amethyst_PointTop,               13);
    CopyBuffer(&Amethyst_Model[i].p[2], &Amethyst_PointMiddle[i],         13);
    CopyBuffer(&Amethyst_Model[i].p[0], &Amethyst_PointMiddle[(i+1) % 6], 13);

    CopyBuffer(&Amethyst_Model[i+6].p[0], &Amethyst_PointMiddleB[i],        13);
    CopyBuffer(&Amethyst_Model[i+6].p[2], &Amethyst_PointMiddle[i],         13);
    CopyBuffer(&Amethyst_Model[i+6].p[1], &Amethyst_PointMiddle[(i+1) % 6], 13);

    CopyBuffer(&Amethyst_Model[i+12].p[0], &Amethyst_PointMiddleB[i],        13);
    CopyBuffer(&Amethyst_Model[i+12].p[1], &Amethyst_PointMiddle[(i+1) % 6],  13);
    CopyBuffer(&Amethyst_Model[i+12].p[2], &Amethyst_PointMiddleB[(i+1) % 6], 13);

    TriangleCalcNormal(&Amethyst_Model[i]);
    TriangleCalcNormal(&Amethyst_Model[i+6]);
    TriangleCalcNormal(&Amethyst_Model[i+12]);
  }
}

static void DrawAmethyst()
{
  int i;
  for(i = 0; i < 18; i++)
    G3D_AddTriangle(&Amethyst_Model[i]);
}

static void RenderAmethyst(unsigned int *Buffer)
{
  G3D_Reset();
  G3D_Translate(-14.0f, 8.0f, 18.0f);
  G3D_ModelFill      = true;

  static float r;
  G3D_RotateAxisY(r+=0.04f);

  DrawAmethyst();

  G3D_Translate(0.0f, 1.5f, -1.2f);
  G3D_RotateAxisX(0.7f);
  G3D_Scale(1.0f, 0.5f, 1.0f);
  DrawAmethyst();

  CurrentMatrix -= 3;
  G3D_Translate(0.0f, 2.2f, 1.1f);
  G3D_RotateAxisX(-0.85f);
  G3D_Scale(0.75f, 0.35f, 1.0f);
  DrawAmethyst();
  
  G3D_Render(Buffer);
}

void EngineRender()
{
  static float r;
  static unsigned int SthBuffer[640*480];
  int i;
  ClearBuffer(BackBuffer, 640*480);

  PShader = PShaderPos;
  VShader = VShaderFlat;
  G3D_Reset();
  G3D_Translate(-65.0f, 45.0f, 0.0f);
  G3D_ModelFill      = false;
  //DrawString3D("\x26\x01\x0d\x05\x14\x08\x19\x13\x14",9);
  DrawString3D("\x08\x00\x04\x01\x06\x02\x07\x05\x06",9);
  G3D_Render(BackBuffer);

  VShader = VShaderIdentity;
  PShader = PShaderColor;
  RenderAmethyst(BackBuffer);

  PShader = PShaderDist;

  G3D_Reset();

  G3D_Translate(0.0f, 0.0f, 58.0f);
  G3D_RotateAxisY(r*0.8f);
  G3D_RotateAxisX(r*2.3f);
  G3D_Translate(-36.0f, 0.0f, 0.0f);
  G3D_ModelFill      = true;
  //DrawString3D("\x01\x13\x08\x09",4);
  DrawString3D("\x00\x05\x02\x03",4);

  r += 0.04f;

  G3D_Render(SthBuffer);  
  
  for(i = 640*479; i > 0 ; i--)
  {
    unsigned int a = 
      *(unsigned char*)&SthBuffer[i] + 
      *(unsigned char*)&SthBuffer[i + 639] +
      *(unsigned char*)&SthBuffer[i + 640] +
      *(unsigned char*)&SthBuffer[i + 641];
    a = (a >> 2)  - my_rand() % 2;
    if(a > 0xff) a = 0;
    *(unsigned char*)&SthBuffer[i] = a;
  }
  
  static double t;
  t += 1.0;

  for(i = 3; i < 640*480 - 3; i++)
  {
    unsigned char p = *(unsigned char*)&SthBuffer[i + (int)(sinf(t + (double)(i / 640) * 0.2f) * 2.0f)];
    BackBuffer[i] |= flame_pal[p];
  }
 
}

